@@ -8,10 +8,31 @@ fn main() {
         .allowlist_type("(_|)kadm5.*")
         .allowlist_function("kadm5.*")
         .allowlist_var("KADM5_.*")
+        // Principal attributes
+        .allowlist_var("KRB5_KDB_DISALLOW_POSTDATED")
+        .allowlist_var("KRB5_KDB_DISALLOW_FORWARDABLE")
+        .allowlist_var("KRB5_KDB_DISALLOW_TGT_BASED")
+        .allowlist_var("KRB5_KDB_DISALLOW_RENEWABLE")
+        .allowlist_var("KRB5_KDB_DISALLOW_PROXIABLE")
+        .allowlist_var("KRB5_KDB_DISALLOW_DUP_SKEY")
+        .allowlist_var("KRB5_KDB_DISALLOW_ALL_TIX")
+        .allowlist_var("KRB5_KDB_REQUIRES_PRE_AUTH")
+        .allowlist_var("KRB5_KDB_REQUIRES_HW_AUTH")
+        .allowlist_var("KRB5_KDB_REQUIRES_PWCHANGE")
+        .allowlist_var("KRB5_KDB_DISALLOW_SVR")
+        .allowlist_var("KRB5_KDB_PWCHANGE_SERVICE")
+        .allowlist_var("KRB5_KDB_SUPPORT_DESMD5")
+        .allowlist_var("KRB5_KDB_NEW_PRINC")
+        .allowlist_var("KRB5_KDB_OK_AS_DELEGATE")
+        .allowlist_var("KRB5_KDB_OK_TO_AUTH_AS_DELEGATE")
+        .allowlist_var("KRB5_KDB_NO_AUTH_DATA_REQUIRED")
+        .allowlist_var("KRB5_KDB_LOCKDOWN_KEYS")
+        // Other utilites
         .allowlist_var("KRB5_NT_SRV_HST")
         .allowlist_var("KRB5_OK")
         .allowlist_var("ENCTYPE_.*")
         .allowlist_var("KRB5_KDB_SALTTYPE_.*")
+        .allowlist_var("KRB5_TL_LAST_ADMIN_UNLOCK")
         .allowlist_function("krb5_init_context")
         .allowlist_function("krb5_free_context")
         .allowlist_function("krb5_get_error_message")
