@@ -7,6 +7,7 @@
 //! The APIs between this wrapper and the underlying [`crate::kadmin::KAdmin`] are the same, and
 //! wrapped and the [`KAdminImpl`] trait.
 use std::{
+    collections::HashMap,
     panic::resume_unwind,
     sync::{
         Arc,
@@ -22,17 +23,38 @@ use crate::{
     db_args::DbArgs,
     error::Result,
     kadmin::{KAdminApiVersion, KAdminImpl},
+    keysalt::KeySalts,
     params::Params,
     policy::{Policy, PolicyBuilder, PolicyModifier},
-    principal::Principal,
+    principal::{Principal, PrincipalBuilder, PrincipalModifier},
 };
 
 /// Operations from [`KAdminImpl`]
 enum KAdminOperation {
+    /// See [`KAdminImpl::add_principal`]
+    AddPrincipal(PrincipalBuilder, Sender<Result<()>>),
+    /// See [`KAdminImpl::modify_principal`]
+    ModifyPrincipal(PrincipalModifier, Sender<Result<()>>),
+    /// See [`KAdminImpl::rename_principal`]
+    RenamePrincipal(String, String, Sender<Result<()>>),
+    /// See [`KAdminImpl::delete_principal`]
+    DeletePrincipal(String, Sender<Result<()>>),
     /// See [`KAdminImpl::get_principal`]
     GetPrincipal(String, Sender<Result<Option<Principal>>>),
     /// See [`KAdminImpl::principal_change_password`]
-    PrincipalChangePassword(String, String, Sender<Result<()>>),
+    PrincipalChangePassword(
+        String,
+        String,
+        Option<bool>,
+        Option<KeySalts>,
+        Sender<Result<()>>,
+    ),
+    /// See [`KAdminImpl::principal_randkey`]
+    PrincipalRandkey(String, Option<bool>, Option<KeySalts>, Sender<Result<()>>),
+    /// See [`KAdminImpl::principal_get_strings`]
+    PrincipalGetStrings(String, Sender<Result<HashMap<String, String>>>),
+    /// See [`KAdminImpl::principal_set_string`]
+    PrincipalSetString(String, String, Option<String>, Sender<Result<()>>),
     /// See [`KAdminImpl::list_principals`]
     ListPrincipals(Option<String>, Sender<Result<Vec<String>>>),
     /// See [`KAdminImpl::add_policy`]
@@ -53,11 +75,37 @@ impl KAdminOperation {
     fn handle(&self, kadmin: &crate::kadmin::KAdmin) {
         match self {
             Self::Exit => (),
+            Self::AddPrincipal(builder, sender) => {
+                let _ = sender.send(kadmin.add_principal(builder));
+            }
+            Self::ModifyPrincipal(modifier, sender) => {
+                let _ = sender.send(kadmin.modify_principal(modifier));
+            }
+            Self::RenamePrincipal(old_name, new_name, sender) => {
+                let _ = sender.send(kadmin.rename_principal(old_name, new_name));
+            }
+            Self::DeletePrincipal(name, sender) => {
+                let _ = sender.send(kadmin.delete_principal(name));
+            }
             Self::GetPrincipal(name, sender) => {
                 let _ = sender.send(kadmin.get_principal(name));
             }
-            Self::PrincipalChangePassword(name, password, sender) => {
-                let _ = sender.send(kadmin.principal_change_password(name, password));
+            Self::PrincipalChangePassword(name, password, keepold, keysalts, sender) => {
+                let _ = sender.send(kadmin.principal_change_password(
+                    name,
+                    password,
+                    *keepold,
+                    keysalts.as_ref(),
+                ));
+            }
+            Self::PrincipalRandkey(name, keepold, keysalts, sender) => {
+                let _ = sender.send(kadmin.principal_randkey(name, *keepold, keysalts.as_ref()));
+            }
+            Self::PrincipalGetStrings(name, sender) => {
+                let _ = sender.send(kadmin.principal_get_strings(name));
+            }
+            Self::PrincipalSetString(name, key, value, sender) => {
+                let _ = sender.send(kadmin.principal_set_string(name, key, value.as_deref()));
             }
             Self::ListPrincipals(query, sender) => {
                 let _ = sender.send(kadmin.list_principals(query.as_deref()));
@@ -117,6 +165,40 @@ impl KAdmin {
 }
 
 impl KAdminImpl for KAdmin {
+    fn add_principal(&self, builder: &PrincipalBuilder) -> Result<()> {
+        let (sender, receiver) = channel();
+        self.inner
+            .op_sender
+            .send(KAdminOperation::AddPrincipal(builder.clone(), sender))?;
+        receiver.recv()?
+    }
+
+    fn modify_principal(&self, modifier: &PrincipalModifier) -> Result<()> {
+        let (sender, receiver) = channel();
+        self.inner
+            .op_sender
+            .send(KAdminOperation::ModifyPrincipal(modifier.clone(), sender))?;
+        receiver.recv()?
+    }
+
+    fn rename_principal(&self, old_name: &str, new_name: &str) -> Result<()> {
+        let (sender, receiver) = channel();
+        self.inner.op_sender.send(KAdminOperation::RenamePrincipal(
+            old_name.to_owned(),
+            new_name.to_owned(),
+            sender,
+        ))?;
+        receiver.recv()?
+    }
+
+    fn delete_principal(&self, name: &str) -> Result<()> {
+        let (sender, receiver) = channel();
+        self.inner
+            .op_sender
+            .send(KAdminOperation::DeletePrincipal(name.to_owned(), sender))?;
+        receiver.recv()?
+    }
+
     fn get_principal(&self, name: &str) -> Result<Option<Principal>> {
         let (sender, receiver) = channel();
         self.inner
@@ -125,13 +207,63 @@ impl KAdminImpl for KAdmin {
         receiver.recv()?
     }
 
-    fn principal_change_password(&self, name: &str, password: &str) -> Result<()> {
+    fn principal_change_password(
+        &self,
+        name: &str,
+        password: &str,
+        keepold: Option<bool>,
+        keysalts: Option<&KeySalts>,
+    ) -> Result<()> {
         let (sender, receiver) = channel();
         self.inner
             .op_sender
             .send(KAdminOperation::PrincipalChangePassword(
                 name.to_owned(),
                 password.to_owned(),
+                keepold,
+                keysalts.cloned(),
+                sender,
+            ))?;
+        receiver.recv()?
+    }
+
+    fn principal_randkey(
+        &self,
+        name: &str,
+        keepold: Option<bool>,
+        keysalts: Option<&KeySalts>,
+    ) -> Result<()> {
+        let (sender, receiver) = channel();
+        self.inner
+            .op_sender
+            .send(KAdminOperation::PrincipalRandkey(
+                name.to_owned(),
+                keepold,
+                keysalts.cloned(),
+                sender,
+            ))?;
+        receiver.recv()?
+    }
+
+    fn principal_get_strings(&self, name: &str) -> Result<HashMap<String, String>> {
+        let (sender, receiver) = channel();
+        self.inner
+            .op_sender
+            .send(KAdminOperation::PrincipalGetStrings(
+                name.to_owned(),
+                sender,
+            ))?;
+        receiver.recv()?
+    }
+
+    fn principal_set_string(&self, name: &str, key: &str, value: Option<&str>) -> Result<()> {
+        let (sender, receiver) = channel();
+        self.inner
+            .op_sender
+            .send(KAdminOperation::PrincipalSetString(
+                name.to_owned(),
+                key.to_owned(),
+                value.map(String::from),
                 sender,
             ))?;
         receiver.recv()?
