@@ -5,9 +5,11 @@ from kadmin._lib import (
     KAdminApiVersion,
     KeySalt,
     KeySalts,
+    NewPrincipalKey,
     Params,
     Policy,
     Principal,
+    PrincipalAttributes,
     SaltType,
     TlData,
     TlDataEntry,
@@ -21,9 +23,11 @@ __all__ = (
     "KAdminApiVersion",
     "KeySalt",
     "KeySalts",
+    "NewPrincipalKey",
     "Params",
     "Policy",
     "Principal",
+    "PrincipalAttributes",
     "SaltType",
     "TlData",
     "TlDataEntry",
